require 'touchgfx/generators/base'
require 'nokogiri'
require 'fileutils'
require 'pathname'

module TouchGFX
  module Generators
    class CubeIDEProjectUpdater
      attr_reader :project_file
      attr_reader :touchgfx_path
      attr_reader :bpp

      # https://help.eclipse.org/luna/topic/org.eclipse.platform.doc.isv/reference/misc/project_description_file.html
      def initialize(project_file, touchgfx_path, touchgfx_env, config, gui_group_name = 'gui', generated_group_name = 'generated')
        @project_file = project_file
        @touchgfx_path = touchgfx_path
        @touchgfx_env = touchgfx_env
        @gui_group_name = gui_group_name
        @generated_group_name = generated_group_name
        @underroot = project_file.include?(".cproject") # Under root uses .cproject, non-under root uses .project
      end

      def run
        @xml = project_file_xml
        update_c_project_file(@xml.xpath('//sourceEntries')) if @underroot
        update_project_file(@xml.at_xpath('//linkedResources'))  if !@underroot

        #########################################
        #   Finalize
        #########################################
        write_project_files

      end

      private

      def update_c_project_file(source_entries)
=begin
         Read entries in <sourceEntries/> - If no entries define "excluding" we need to add one for each folder
         following the following rules:

           1) Middlewares/ST/TouchGFX - simulator files
           2) TouchGFX/               - simulator files
           3) CubeMX                  - We need to exclude OSWrapper_cmsis.cpp
                                        _cmsis.cpp won't support V2, so we'll fall back to freertos native OSWrapper.cpp for now.
=end
        Dir.chdir("..") do |d|
          puts "Now in #{d}. #{Dir.pwd}"

          #########################################
          #   Create exclude entries for TouchGFX
          #   and Middlewares folders
          #########################################

          mw_files_to_exclude = Dir["Middlewares/**/TouchGFX/**/simulator/*/*.cpp"]
          additional_files = ["ST/TouchGFX/touchgfx/os/OSWrappers_cmsis.cpp", "ST/TouchGFX/touchgfx/framework/source/platform/driver/touch/SDL2TouchController.cpp", "ST/TouchGFX/touchgfx_backup*"]
          mw_files_to_exclude.map! { |e| e[12..-1] }
          mw_exclude_entry = "#{mw_files_to_exclude.join("|")}|#{additional_files.join("|")}"

          app_files_to_exclude = Dir["TouchGFX/**/simulator/**/*.cpp"]
          app_files_to_exclude.map! {|e| e[9..-1]}
          app_exclude_entry = "#{app_files_to_exclude.join("|")}"

          source_entries.each do |a|
            a.children.each do |b|
              update_excluding(b, mw_exclude_entry) if b["name"] == "Middlewares"
              update_excluding(b, app_exclude_entry) if b["name"] == "TouchGFX"
            end
          end
        end
      end

      def update_project_file(source_entries)
        prefix_folder = Pathname.new(Dir.getwd).basename.to_s
        entry_list = source_entries.children.map {|c| c.inner_text}.join
        [
          [@gui_group_name, 'gui/**/*.{cpp,c}'],
          [@generated_group_name, 'generated/{fonts,gui_generated,images,texts}/**/*.{cpp,c}']
        ].each do |group_name, pattern|
          Dir.glob(pattern).each do |file_name|
            if !entry_list.include?(file_name)
              add_linkedResources_entry(["Application", "User", group_name].join('/'), [prefix_folder, file_name].join('/'), source_entries)
            end
          end
        end
      end

      def add_linkedResources_entry(group, filename, source_entries)
        link = create_node("link",nil, source_entries)
        create_node("name", [group, File.basename(filename)].join('/'), link)
        create_node("type", 1 , link)
        create_node("locationURI", ["$%7BPARENT-1-PROJECT_LOC%7D", filename].join('/'), link)
      end

      def create_node(type, content, parent)
        begin
          node = Nokogiri::XML::Node.new type, project_file_xml
          node.parent = parent
          node.content = content if content
          parent.add_child node
        end
      end

      def update_excluding group, entry
        # Updating Excluding
        current_excluding_list = group['excluding'] ? group['excluding'].split("|") : Array.new
        uniq_excluding_entries = (current_excluding_list + entry.split("|")).uniq.join("|")
        group['excluding'] = uniq_excluding_entries

        # Updating Flags
        current_flags_list = group['flags'] ? group['flags'].split("|") : Array.new
        uniq_flags_entries = (current_flags_list + ["VALUE_WORKSPACE_PATH"]).uniq.join("|")
        group['flags'] = uniq_flags_entries
      end

      def add_excluding group, entry
        group['excluding'] = entry
        group['flags'] = "VALUE_WORKSPACE_PATH"
      end

      def project_file_contents
        File.read(project_file, :encoding=>'utf-8')
      end

      def project_file_xml
        @project_file_xml ||= Nokogiri::XML(project_file_contents) do |config|
          config.default_xml.noblanks
        end
      end

      def write_project_files
          output = @xml.to_xml(:indent => 2)
          output.gsub!(/  /, "\t") if !@underroot
          File.open(@project_file, 'w') { |f| f.write(output) }
          FileUtils.touch('../.project')
      end
    end
  end
end
