require 'touchgfx/generators/base'
require 'nokogiri'
require 'pathname'

module TouchGFX
  module Generators
    class IAR < Base

      argument :project_file
      argument :bpp, :optional => true
      argument :only_project, :optional => true
      PROJ_DIR = '$PROJ_DIR$/../../../..'

      def check_project_node
        if project_file_xml.at_xpath('./project').nil?
          say_status :conflict, "no <project> node in #{project_file}", :red # #{project_file}
          Kernel.abort
        end
      end

      def project_file_contents
        @tgfx_path = "#{ENV['touchgfx'].strip}/touchgfx"
        if project_file.match('/touchgfx\.ewp')
          update_touchgfx
        else
          ensure_source_nodes
          output_project_file
        end
      end

      private

      def update_touchgfx
        level = project_file.split('/').size() - 3

        touchgfx_dir = "#{'../'*level}"

        project_node = project_file_xml.at_xpath('./project')
        touchgfx_node = group_with_name(project_node, 'TouchGFX')
        clean_group(touchgfx_node)

        Dir["touchgfx/framework/source/common/**/*.cpp",
            "touchgfx/framework/source/platform/driver/lcd/**/*.cpp",
            "touchgfx/framework/source/touchgfx/**/*.cpp"].each do |file_name|
          path = file_name.split('/')[3..-2]
          file_with_file_path(touchgfx_node, "#{touchgfx_dir}#{file_name}", path)
        end

        output = project_file_xml.
                   to_xml(:indent => 2).
                   gsub('..\\..\\..\\touchgfx\\','').
                   gsub(/<(.*?)\/>/, '<\\1></\\1>').
                   gsub('<data></data>', '<data/>').
                   gsub('<project>',"\n<project>").
                   gsub('</project>',"</project>\n\n")
        write_file(project_file, output)
      end

      def ensure_source_nodes
        @project_node = project_file_xml.at_xpath('./project')

        @application_node = group_with_name(@project_node,'Application')
        # 'gui','generated' and 'main' are the only folders generated by the code after this "each". Remove those groups.
        application_groups = ['gui','generated']
        unless only_project
          application_groups << 'main'
        end
        application_groups.each do |group|
          existing_group = group_with_name(@application_node, group, true)
          if existing_group
            clean_group(existing_group)
          end
        end
        patterns = ['gui/src/**/*.cpp',
                    'gui/src/**/*.c',
                    'generated/images/**/*.cpp',
                    'generated/texts/**/*.cpp',
                    'generated/fonts/**/*.cpp',
                    'generated/gui_generated/**/*.cpp']
        unless only_project
          patterns += ['target/src/**/*.cpp',
                       'target/src/**/*.c',
                       'target/*.cpp']
        end
        Dir.glob(patterns).sort_by(&iar_file_order).each do |file_name|
          path = file_name.gsub('/src/','/').gsub(/target\//,'main/').split('/')[0..-2]
          file_with_file_path(@application_node, file_name, path, true)
        end

        unless only_project
          library = @project_node.at_xpath('./configuration/settings/data/option[name="IlinkAdditionalLibs"]/state').text
          if library.empty?
            # Scan touchgfx folder for relevant .cpp files
            touchgfx_dir = "#{@tgfx_path}/framework/source/"
            files = [Dir["#{touchgfx_dir}common/**/*.cpp",
                         "#{touchgfx_dir}platform/driver/lcd/**/*.cpp",
                         "#{touchgfx_dir}touchgfx/**/*.cpp"].sort_by {|f| File.basename(f).downcase }].flatten
            touchgfx_node = group_with_name(@project_node,'TouchGFX')
            clean_group(touchgfx_node)

            level = @tgfx_path.split('\\').size
            files.each do |file_name|
              path = file_name[touchgfx_dir.length..-1].split('/')[0..-2] # remove touchgfx_dir from start and filename from ending
              file_with_file_path(touchgfx_node, file_name, path)
            end
          end
        end

        unless only_project
          # Fixup touchgfx relative paths
          @project_node.xpath("//*[contains(text(),'\\touchgfx\\')]").each { |node|
            old_touchgfx_file = node.content.to_s
            new_touchgfx_file = File.join(@tgfx_path,old_touchgfx_file.match(/\\touchgfx\\(.*)/)[1]).gsub('/','\\')
            if File.exist?(new_touchgfx_file)
              node.content = relative_path(new_touchgfx_file).gsub(/\\\\+/,'\\')
            end
          }
        end

        states = @project_node.at_xpath('./configuration/settings[name="ICCARM"]/data/option[name="CCDefines"]')

        # Read existing USE_BPP
        use_bpp_state = states.at_xpath('state[starts-with(.,"USE_BPP")]')
        old_bpp = 16 # default
        if use_bpp_state
          old_bpp = use_bpp_state.content.match(/USE_BPP=(\d+)/)[1].to_i
          old_bpp = 16 if old_bpp == 0
        end

        new_bpp = bpp == 18 ? 24 : bpp
        if Dir.pwd.match(/_24bit$/)
          new_bpp = 24
        elsif Dir.pwd.match(/_4bit$/)
          new_bpp = 4
        elsif new_bpp == 0
          new_bpp = old_bpp
        end

        if new_bpp != old_bpp
          while states.at_xpath('state[starts-with(.,"USE_BPP")]')
            states.at_xpath('state[starts-with(.,"USE_BPP")]').remove()
          end
          state_use_bpp = Nokogiri::XML::Node.new 'state', project_file_xml
          state_use_bpp.content = "USE_BPP=#{new_bpp}"
          state_use_bpp.parent = states
          states.add_child(state_use_bpp)
        end
      end

      def tgfx_rel_path
        unless @tgfx_rel_path
          tgfx_array = @tgfx_path.gsub(/^(.:)/){$1.downcase}.gsub('\\','/').split('/')
          if @tgfx_path[0] == '.'
            @tgfx_rel_path = File.join(tgfx_array)
          else
            cur_array = Dir.pwd.gsub(/^(.:)/){$1.downcase}.gsub('\\','/').split('/')
            while tgfx_array[0] == cur_array[0]
              tgfx_array = tgfx_array[1..-1]
              cur_array = cur_array[1..-1]
            end
            @tgfx_rel_path = File.join(['..']*cur_array.size, tgfx_array)
          end
          if PROJ_DIR.length > 0
            @tgfx_rel_path = File.join(PROJ_DIR, @tgfx_rel_path)
          end
          if @tgfx_rel_path.match(/:/)
            @tgfx_rel_path = @tgfx_path
          end
          @tgfx_rel_path.gsub!('/','\\')
        end
        return @tgfx_rel_path
      end

      def nice_path(path)
        path.gsub!('/','\\')
        path = path[0..-2] if path[-1]=='\\'
        return path
      end

      def relative_path(file_path)
        cur_path = Dir.pwd
        proj_path = PROJ_DIR
        file_path.gsub!('\\','/')
        if file_path.start_with?(@tgfx_path)
          return nice_path(File.join(tgfx_rel_path, file_path[@tgfx_path.length..-1]))
        elsif file_path.start_with?(cur_path)
          return nice_path(File.join(proj_path, file_path[cur_path.length..-1]))
        elsif file_path.match(/:/)
          return nice_path(file_path)
        end
        if only_project
          ewp_strpath = Pathname.new(File.dirname(project_file)).realpath
          file_strpath = Pathname.new(file_path).realpath
          relative_path = file_strpath.relative_path_from(ewp_strpath).to_s
          return nice_path(File.join('$PROJ_DIR$', relative_path))
        end
        return nice_path(File.join(proj_path, file_path))
      end

      def write_file(file, output)
        input = File.read(file, :encoding=>'utf-8')
        if input != output
          File.open(file, 'w:utf-8') { |f| f.write(output) }
        end
      end

      def clean_group(node)
        while (group = node.at_xpath('./group'))
          group.remove()
        end
        while (file = node.at_xpath('./file'))
          file.remove()
        end
      end

      def group_with_name(parent, name, anywhere=false)
        element_with_type_name_and_contents(parent, anywhere, 'group', name, name)
      end

      def option_with_name(parent, name)
        element_with_type_name_and_contents(parent, false, 'option', name, name)
      end

      def file_with_file_path(parent, file_name, path, anywhere=false)
        if path.size == 0
          file_with_file_name(parent, file_name)
        else
          node = group_with_name(parent, path[0], anywhere)
          file_with_file_path(node, file_name, path[1..-1], false)
        end
      end

      def file_with_file_name(parent, file_name)
        path = relative_path(file_name).gsub('..\\target\\','')
        element_with_type_name_and_contents(parent, false, 'file', file_name, path)
      end

      def element_with_type_name_and_contents(parent, anywhere, type, name, contents)
        parent.at_xpath(".#{anywhere ? '//' : '/'}#{type}[name=\"#{contents}\"]") ||
          begin
            # say_status :create, "#{type} #{name}", :green
            node = Nokogiri::XML::Node.new type, project_file_xml
            name_node = Nokogiri::XML::Node.new 'name', project_file_xml
            name_node.content = contents
            name_node.parent = node

            parent.add_child node
          end
      end

      def target
        File.join(Dir.pwd)
      end

      def project_file_xml
        @project_file_xml ||= Nokogiri::XML(File.read(project_file, :encoding=>'utf-8')) do |config|
          config.default_xml.noblanks
        end
      end

      def output_project_file
        output = project_file_xml.to_xml(:indent => 2)
#        empty_terms = ["state", "extensions", "cmdline", "prebuild", "postbuild"]
#        output = empty_terms.inject(output) do |result, term|
#          result.gsub("<#{term}/>", "<#{term}></#{term}>")
#        end
#        output = (output.lines.to_a.insert(1,"\n") << "\n" << "\n").join
        File.open(project_file, 'wb') { |file| file.write(output) }
      end

      def iar_file_order
        proc { |word| word.downcase }
      end
    end
  end
end
