require 'touchgfx/generators/base'
require 'nokogiri'
require 'pathname'

module TouchGFX
  module Generators
    class IARProjectConnector
      attr_reader :project_file
      attr_reader :touchgfx_path
      attr_reader :bpp

      def initialize(project_file, touchgfx_path, touchgfx_env, config, gui_group_name = 'gui', generated_group_name = 'generated')
        @project_file = project_file
        @touchgfx_path = touchgfx_path
        @touchgfx_env = touchgfx_env
        @bpp = config.bpp
        @gui_group_name = gui_group_name
        @generated_group_name = generated_group_name
      end

      def run
        project_connection = create_IPCF_file
        output_project_file(project_connection)
        ensure_usage_of_IPCF_file
      end

      def create_IPCF_file
        Nokogiri::XML::Builder.new do |xml|
          xml.iarProjectConnection('name' => 'TouchGFX', 'oneShot' => 'false') do
            xml.includePath { update_includes([@generated_group_name, @gui_group_name]).each { |p| xml.path p } }
            xml.asmIncludePath
            xml.defines {xml.define "USE_BPP=#{@bpp}" }
            xml.asmDefines
            xml.files { [@generated_group_name, @gui_group_name].each { |file_group|
              xml.group('name' => file_group) { find_files("#{file_group}/**/*.{cpp,c}").each { |group|
                xml.group('name' => group.first[0]) { group.last.each {|f| xml.path f } } } } } }
          end
        end
      end

      def project_file_contents
        File.read(project_file, :encoding=>'utf-8') if File.exists?(project_file)
      end

      def find_files(pattern)
        files = Dir.glob(pattern).sort_by(&iar_file_order)
        commonPath = common_directory_path(files)

        d = {}
        files.each { |f| group = f.match(/^#{commonPath}\/(\w+)/).captures
          (!d.key? group )? d[group] = [path_to(f)] : d[group] << path_to(f)
        }
        return d
      end

      def common_directory_path(dirs, separator='/')
        dir1, dir2 = dirs.minmax.map{|dir| dir.split(separator)}
        dir1.zip(dir2).take_while{|dn1,dn2| dn1==dn2}.map(&:first).join(separator)
      end

      def update_includes(include_base_paths)
        list = [@touchgfx_path + "framework/include"]
        list << Dir.glob("{#{include_base_paths.join(',')}}/**/include").sort_by(&iar_file_order)
        return list.flatten.map {|d| path_to(d)}
      end

      def path_to(file_path)
        absolute_file_path = File.absolute_path(file_path, Dir.pwd)
        project_dir = File.dirname(project_file)
        absolute_project_dir = File.absolute_path(project_dir, Dir.pwd)
        relative_path = Pathname.new(absolute_file_path).relative_path_from(Pathname.new(absolute_project_dir))
        return File.join("$PROJ_DIR$", relative_path).gsub('/', '\\')
      end

      def iar_file_order
        proc { |word| word.downcase }
      end

      def ensure_usage_of_IPCF_file
        usingIPCF = Dir.glob(File.dirname(@project_file) + "/*.ewp").sort_by(&iar_file_order).map { |f|
            File.read(f, :encoding=>'utf-8').match(File.basename(@project_file))
        }.any?

        abort("Please add \"#{@project_file}\" to IAR under \"Project->Add Project Connection->IAR Project Connection\"") if !usingIPCF
      end

      def output_project_file(xml)
        output = xml.to_xml(:indent => 2,
          :save_with => Nokogiri::XML::Node::SaveOptions::NO_EMPTY_TAGS | Nokogiri::XML::Node::SaveOptions::FORMAT)
        if(output != project_file_contents)
          File.open(project_file, 'wb') { |file| file.write(output) }
        end
      end
    end
  end
end