require 'touchgfx/generators/base'
require 'nokogiri'
require 'pathname'

module TouchGFX
  module Generators
    class IARProjectUpdater
      attr_reader :project_file
      attr_reader :touchgfx_path
      attr_reader :bpp

      def initialize(project_file, touchgfx_path, touchgfx_env, config, gui_group_name = 'gui', generated_group_name = 'generated')
        @project_file = project_file
        @touchgfx_path = touchgfx_path
        @touchgfx_env = touchgfx_env
        @bpp = config.bpp
        @gui_group_name = gui_group_name
        @generated_group_name = generated_group_name
      end

      def run
        check_project_node
        ensure_source_nodes
        update_includes
        update_defines
        update_library
        output_project_file
      end

      private

      def check_project_node
        project_file_xml.at_xpath('./project') ||
          abort("no <project> node in #{project_file}")
      end

      def ensure_source_nodes
        @project_node = project_file_xml.at_xpath('./project')

        [@gui_group_name, @generated_group_name].each do |name|
          group = group_with_name(@project_node, name, true)
          clean_group(group)
        end
        [
          [@gui_group_name, 'gui/**/*.{cpp,c}'],
          [@generated_group_name, 'generated/**/*.{cpp,c}']
        ].each do |group_name, pattern|
          Dir.glob(pattern).sort_by(&iar_file_order).each do |file_name|
            path = file_name.gsub('/src/','/').split('/')[1..-2]
            group = group_with_name(@project_node, group_name, true)
            file_with_file_path(group, file_name, path, true)
          end
        end
      end

      def update_includes
        include_path = 'touchgfx\framework\include'
        state_includes = [include_path, include_path.gsub("\\","/")].collect do |path|
          @project_node.xpath(".//state[substring(., string-length(.) - string-length('#{path}') + 1) = '#{path}']") #ends-with does not work in XPath 1.0
        end.flatten
        abort("no include of #{include_path} found in #{project_file}") if state_includes.empty?
        state_includes.each { |s| s.content = path_to(File.join(touchgfx_path, 'framework', 'include')) }
      end

      def update_defines
        states = @project_node.at_xpath('./configuration/settings[name="ICCARM"]/data/option[name="CCDefines"]')
        state_use_bpp = states.at_xpath('state[starts-with(.,"USE_BPP")]') ||
          begin
            state = Nokogiri::XML::Node.new 'state', project_file_xml
            state.parent = states
            states.add_child state
          end
        state_use_bpp.content = "USE_BPP=#{bpp}"
      end

      def update_library
        #library can be present in states or files
        libraries = '.a'
        states = @project_node.xpath(".//state[substring(., string-length(.) - string-length('#{libraries}') + 1) = '#{libraries}']") #ends-with does not work in XPath 1.0
        states = states.select { |s| s.content.match /touchgfx[\\\/]lib/ }
        names = @project_node.xpath(".//name[substring(., string-length(.) - string-length('#{libraries}') + 1) = '#{libraries}']") #ends-with does not work in XPath 1.0
        names = names.select { |s| s.content.match /touchgfx[\\\/]lib/ }
        abort("no touchgfx library found in #{project_file}") if states.empty? && names.empty?
        states.each { |s|
          s.content = s.content.gsub(/.*touchgfx[\\\/]lib/, "#{path_to(touchgfx_path)}\\lib".gsub('\\','/')).gsub('/','\\') # backslashes are evil, as they will escape 4 in \\4.8.0
        }
        names.each { |s|
          s.content = s.content.gsub(/.*touchgfx[\\\/]lib/, "#{path_to(touchgfx_path)}\\lib".gsub('\\','/')).gsub('/','\\') # backslashes are evil, as they will escape 4 in \\4.8.0
        }
      end

      def path_to(file_path)
        absolute_file_path = File.absolute_path(file_path, Dir.pwd)
        project_dir = File.dirname(project_file)
        absolute_project_dir = File.absolute_path(project_dir, Dir.pwd)
        relative_path = Pathname.new(absolute_file_path).relative_path_from(Pathname.new(absolute_project_dir))
        return File.join("$PROJ_DIR$", relative_path).gsub('/', '\\')
      end

      def clean_group(node)
        while (group = node.at_xpath('./group'))
          group.remove()
        end
        while (file = node.at_xpath('./file'))
          file.remove()
        end
      end

      def group_with_name(parent, name, anywhere=false)
        element_with_type_name_and_contents(parent, anywhere, 'group', name, name)
      end

      def option_with_name(parent, name)
        element_with_type_name_and_contents(parent, false, 'option', name, name)
      end

      def file_with_file_path(parent, file_name, path, anywhere=false)
        if path.size == 0
          file_with_file_name(parent, file_name)
        else
          node = group_with_name(parent, path[0], anywhere)
          file_with_file_path(node, file_name, path[1..-1], false)
        end
      end

      def file_with_file_name(parent, file_name)
        path = path_to(file_name)
        element_with_type_name_and_contents(parent, false, 'file', file_name, path)
      end

      def element_with_type_name_and_contents(parent, anywhere, type, name, contents)
        parent.at_xpath(".#{anywhere ? '//' : '/'}#{type}[name=\"#{contents}\"]") ||
        begin
          node = Nokogiri::XML::Node.new type, project_file_xml
          name_node = Nokogiri::XML::Node.new 'name', project_file_xml
          name_node.content = contents
          name_node.parent = node

          parent.add_child node
        end
      end

      def project_file_contents
        File.read(project_file, :encoding=>'utf-8')
      end

      def project_file_xml
        @project_file_xml ||= Nokogiri::XML(project_file_contents) do |config|
          config.default_xml.noblanks
        end
      end

      def output_project_file
        output = project_file_xml.to_xml(:indent => 2,
          :save_with => Nokogiri::XML::Node::SaveOptions::NO_EMPTY_TAGS | Nokogiri::XML::Node::SaveOptions::FORMAT)
        if(output != project_file_contents)
          File.open(project_file, 'wb') { |file| file.write(output) }
        end
      end

      def iar_file_order
        proc { |word| word.downcase }
      end

      def abort(msg)
        STDERR.puts msg; Kernel.abort
      end
    end
  end
end
