require 'touchgfx/generators/base'
require 'nokogiri'

module TouchGFX
  module Generators
    class Keil < Base

      argument :project_file
      argument :bpp, :optional => true
      argument :only_project, :optional => true
      PROJ_DIR = '..\..\..\..'

      def check_project_node
        if project_file_xml.at_xpath('./Project').nil?
          say_status :conflict, "no <project> node in #{project_file}", :red # #{project_file}
          Kernel.abort
        end
      end

      def project_file_contents
        @tgfx_path = "#{ENV['touchgfx'].strip}/touchgfx"
        if project_file.match('/touchgfx.uvproj')
          update_touchgfx
        else
          update_project
          output_project_file
        end
      end

      private

      def update_touchgfx
        @project_node = project_file_xml.at_xpath('./Project') #.root

        level = project_file.split('/').size() - 2
        touchgfx_dir = "#{'../'*level}"

        [['common', 'framework/source/common'],
         ['platform', 'framework/source/platform/driver/lcd'],
         ['touchgfx', 'framework/source/touchgfx']].each do |groupname, folder|

          #Locate group node
          group_node = project_file_xml.xpath("//Groups/Group[GroupName='#{groupname}']") #There will be nr of nodes depending on configurations. refactor to group_node

          group_node.each do | node|
            files_node = node.at_xpath("Files")   #Get fixpoint for adding file nodes

            files_node.children.remove

            #Create file nodes
            Dir["touchgfx/#{folder}/**/*.cpp"].sort_by { |path| File.basename(path).downcase }.each do |file_name|
              create_file_node(files_node, "#{touchgfx_dir}#{file_name}", 8)
            end
            Dir["touchgfx/#{folder}/**/*.txt"].sort_by { |path| File.basename(path).downcase }.each do |file_name|
              create_file_node(files_node, "#{touchgfx_dir}#{file_name}", 5)
            end
          end

        end

        # Fixup touchgfx relative paths
        unless only_project
          @project_node.xpath("//*[contains(text(),'\\touchgfx\\')]").each do |node|
            content = node.content.to_s
            node.content = content.split(';').map do |path|
              if path.match(/\\touchgfx\\/)
                path = File.join(touchgfx_dir,path.match(/\\touchgfx\\(.*)/)[1]).gsub('/','\\')
              end
              path
            end.join(';')
          end
        end

        write_xml_to_file(project_file_xml.to_xml(:indent => 2)
                           .gsub('..\\..\\..\\touchgfx\\',''), project_file)
      end

      def update_project
        @project_node = project_file_xml.at_xpath('./Project') #.root

        groups = [
          ['Gui',          Dir['gui/src/**/*.cpp','gui/src/**/*.c'].sort_by {|f| File.basename(f).downcase }],
          ['Generated',    Dir['generated/images/**/*.cpp',
                               'generated/texts/**/*.cpp',
                               'generated/fonts/**/*.cpp'].sort_by {|f| File.basename(f).downcase }],
          ['GuiGenerated', Dir['generated/gui_generated/**/*.cpp'].sort_by {|f| File.basename(f).downcase }]
        ]
        unless only_project
          groups += [ ['Target', Dir['target/src/**/*.cpp','target/src/**/*.c'].sort_by {|f| File.basename(f).downcase }] ]
        end
        @project_node.xpath("//Group/GroupName").each do |groupname|
          groupname.content = 'Lib' if groupname.content.match(/^lib$/i)
          groupname.content = 'TouchGFX' if groupname.content.match(/^touchgfxcore$/i)
          groupname.content = 'TouchGFX' if groupname.content.match(/^touchgfx$/i)
        end

        unless only_project
          library = @project_node.xpath("//Group[GroupName='Lib']/Files/File/FileName").text
          if library.match(/^touchgfx_core\.lib$/)
            touchgfx_group = @project_node.xpath("//Group[GroupName='TouchGFX']")
            touchgfx_group.remove if touchgfx_group
          else
            # Scan touchgfx folder for relevant .cpp files
            groups << ['TouchGFX',
                       [Dir["#{@tgfx_path}/framework/source/common/**/*.cpp",
                            "#{@tgfx_path}/framework/source/platform/driver/lcd/**/*.cpp",
                            "#{@tgfx_path}/framework/source/touchgfx/**/*.cpp",
                            "#{@tgfx_path}/framework/source/touchgfx/**/*.txt"].sort_by {|f| File.basename(f).downcase }].flatten]
          end
        end

        groups.each do |groupname, files|
          unless files.empty?
            #Locate group node
            group_node = @project_node.xpath("//Group[GroupName='#{groupname}']") #There will be nr of nodes depending on configurations. refactor to group_node

            if group_node.empty?
              groups_node = @project_node.xpath("//Groups")[0]
              group_node = Nokogiri::XML::Node.new "Group", project_file_xml
              group_node.parent = groups_node
              groupname_node = Nokogiri::XML::Node.new "GroupName", project_file_xml
              groupname_node.parent = group_node
              groupname_node.content = groupname
              files_node = Nokogiri::XML::Node.new "Files", project_file_xml
              files_node.parent = group_node

              group_node = @project_node.xpath("//Group[GroupName='#{groupname}']")
            end

            #Depending on Nr of project-configurations, we may have more than one group that needs to be updated.
            group_node.each do | node|
              files_node = node.at_xpath("Files")   #Get fixpoint for adding file nodes

              files_node.children.remove

              #Create file nodes
              files.each do |file_name|
                create_file_node(files_node, file_name, file_name.match(/\.c$/) ? 1 : file_name.match(/\.txt$/) ? 5 : 8)
              end
            end
          end
        end

        # Try to create this order:
        # ..., Gui, GuiGenerated, Generated, ...
        gui_node = @project_node.xpath("//Group[GroupName='Gui']")
        if !gui_node.empty?
          generated_node = @project_node.xpath("//Group[GroupName='Generated']")
          if !generated_node.empty?
            gui_node.last.next = generated_node
          end
          gui_generated_node = @project_node.xpath("//Group[GroupName='GuiGenerated']")
          if !gui_generated_node.empty?
            gui_node.last.next = gui_generated_node
          end
        end

        # Fixup touchgfx relative paths
        @project_node.xpath("//*[contains(text(),'\\touchgfx\\')]").each do |node|
          content = node.content.to_s
          node.content = content.split(';').map do |path|
            if path.match(/\\touchgfx\\/)
              new_path = File.join(@tgfx_path,path.match(/\\touchgfx\\(.*)/)[1]).gsub('/','\\')
              if File.exist?(new_path)
                path = relative_path(new_path).gsub(/\\\\+/,'\\')
              end
            end
            path
          end.join(';')
        end

        defines=@project_node.xpath('//Cads/VariousControls/Define')
        new_bpp = bpp == 18 ? 24 : bpp
        if new_bpp == 0
          # Use exising bpp, read it from file
          old_bpp = defines[0].content.split(' ').find { |define| define.start_with?('USE_BPP=') }
          if old_bpp
            new_bpp = old_bpp.match(/USE_BPP=(\d+)/)[1].to_i
            new_bpp = 16 if new_bpp == 0
          else
            new_bpp = 16
          end
        end
        if Dir.pwd.match(/_24bit$/)
          new_bpp = 24
        end
        if Dir.pwd.match(/_4bit$/)
          new_bpp = 4
        end
        defines[0].content = ( defines[0].content.split(' ').reject { |define| define.start_with?('USE_BPP=') } + ["USE_BPP=#{new_bpp}"] ) * ' '

      end

      def create_file_node(parent_node, file_name, file_type)
        file_node = Nokogiri::XML::Node.new "File", project_file_xml
        path = relative_path(file_name.gsub('/','\\')).gsub("..\\target\\", "")

        [
          ["FileName", File.basename(file_name)],
          ["FileType", file_type],
          ["FilePath", path]
        ].each do |node_name, node_content|
           create_file_node_content(node_name, node_content, file_node)
        end

        file_node.parent = parent_node
      end

      def create_file_node_content(node_name, node_content, node_parent)
        filename = Nokogiri::XML::Node.new "#{node_name}", project_file_xml
        filename.content = node_content
        filename.parent = node_parent
      end

      ### ALSO IN IAR.RB ###
      def tgfx_rel_path
        unless @tgfx_rel_path
          tgfx_array = @tgfx_path.gsub(/^(.:)/){$1.downcase}.gsub('\\','/').split('/')
          if @tgfx_path[0] == '.'
            @tgfx_rel_path = File.join(tgfx_array)
          else
            cur_array = Dir.pwd.gsub(/^(.:)/){$1.downcase}.gsub('\\','/').split('/')
            while tgfx_array[0] == cur_array[0]
              tgfx_array = tgfx_array[1..-1]
              cur_array = cur_array[1..-1]
            end
            @tgfx_rel_path = File.join(['..']*cur_array.size, tgfx_array)
          end
          if PROJ_DIR.length > 0
            @tgfx_rel_path = File.join(PROJ_DIR, @tgfx_rel_path)
          end
          if @tgfx_rel_path.match(/:/)
            @tgfx_rel_path = @tgfx_path
          end
          @tgfx_rel_path.gsub!('/','\\')
        end
        return @tgfx_rel_path
      end

      def nice_path(path)
        path.gsub!('/','\\')
        path = path[0..-2] if path[-1]=='\\'
        return path
      end

      def relative_path(file_path)
        cur_path = Dir.pwd
        proj_path = PROJ_DIR

        if file_path.start_with?(@tgfx_path.gsub('/','\\'))
          return nice_path(File.join(tgfx_rel_path, file_path[@tgfx_path.length..-1]))
        elsif file_path.start_with?(cur_path)
          return nice_path(File.join(proj_path, file_path[cur_path.length..-1]))
        elsif file_path.match(/:/)
          return nice_path(file_path)
        end
        if only_project
          ewp_strpath = Pathname.new(File.dirname(project_file)).realpath
          file_strpath = Pathname.new(file_path).realpath
          relative_path = file_strpath.relative_path_from(ewp_strpath).to_s
          return nice_path(relative_path)
        end
        return nice_path(File.join(proj_path, file_path))
      end

      def write_file(file, output)
        input = File.read(file, :encoding=>'utf-8')
        if input != output
          File.open(file, 'w:utf-8') { |f| f.write(output) }
        end
      end

      def project_file_xml
        @project_file_xml ||= Nokogiri::XML(File.read(project_file, :encoding=>'utf-8')) do |config|
          config.default_xml.noblanks
        end
      end

      def output_project_file
        write_xml_to_file(project_file_xml.to_xml(:indent => 2), project_file)
      end

      def write_xml_to_file(xml,file)
        # convert <foo/> to <foo></foo>
        output = xml.
                 gsub(/<(.*?)\/>/, '<\\1></\\1>').
                 gsub('?>', ' ?>').
                 gsub('  <SchemaVersion>', "\n  <SchemaVersion>").
                 gsub('  <Header>', "\n  <Header>").
                 gsub('  <Targets>', "\n  <Targets>").
                 gsub('</Project>', "\n</Project>")
        write_file(project_file, output)
      end
    end
  end
end
