require 'touchgfx/generators/base'
require 'nokogiri'
require 'pathname'

module TouchGFX
  module Generators
    class KeilProjectUpdater
      attr_reader :project_file
      attr_reader :touchgfx_path
      attr_reader :bpp

      def initialize(project_file, touchgfx_path, touchgfx_env, config, gui_group_name = 'gui', generated_group_name = 'generated')
        @project_file = project_file
        @touchgfx_path = touchgfx_path
        @touchgfx_env = touchgfx_env
        @bpp = config.bpp
        @gui_group_name = gui_group_name
        @generated_group_name = generated_group_name
      end

      def run
        check_project_node
        ensure_source_nodes
        update_includes
        update_defines
        update_library
        output_project_file
      end

      private

      def check_project_node
        project_file_xml.at_xpath('./Project') ||
        abort("no <Project> node in #{project_file}")
      end

      def ensure_source_nodes
        @project_node = project_file_xml.at_xpath('./Project')

        [@gui_group_name, @generated_group_name].each do |name|
          group = @project_node.at_xpath("//Group[GroupName='#{name}']")
          group ||=
          begin
            groups_node = @project_node.xpath("//Groups")[0]
            group_node = Nokogiri::XML::Node.new "Group", project_file_xml
            group_node.parent = groups_node
            groupname_node = Nokogiri::XML::Node.new "GroupName", project_file_xml
            groupname_node.parent = group_node
            groupname_node.content = name
            files_node = Nokogiri::XML::Node.new "Files", project_file_xml
            files_node.parent = group_node
            group_node
          end
          puts group.xpath("./Files/File")
          group.xpath("./Files/File").remove
        end
        [
          [@gui_group_name, 'gui/**/*.{cpp,c}'],
          [@generated_group_name, 'generated/**/*.{cpp,c}']
        ].each do |group_name, pattern|
          file_paths = Dir.glob(pattern).sort_by(&keil_file_order)
          groups_with_name = @project_node.xpath("//Group[GroupName='#{group_name}']")
          groups_with_name.each do |group|
            files_node = group.at_xpath("./Files")
            file_paths.each do |file_name|
              file_node = Nokogiri::XML::Node.new "File", project_file_xml
              file_node.parent = files_node
              file_name_node = Nokogiri::XML::Node.new( "FileName", project_file_xml ).
                tap { |f| f.parent = file_node; f.content = File.basename(file_name) }
              file_type_node = Nokogiri::XML::Node.new( "FileType", project_file_xml ).
                tap { |f| f.parent = file_node; f.content = file_name.match(/\.c$/) ? 1 : 8 }
              file_path_node = Nokogiri::XML::Node.new( "FilePath", project_file_xml ).
                tap { |f| f.parent = file_node; f.content = path_to(file_name) }
            end

          end
        end
      end

      def update_includes
        includes = @project_node.xpath("//IncludePath")
        includes.each do |include|
          include.content = include.content.split(";").
            collect { |i| i.gsub(/^.*?touchgfx[\\\/]framework[\\\/]include/, path_to(File.join(touchgfx_path, 'framework', 'include')))}.
            join(";")
        end
      end

      def update_defines
        defines = @project_node.xpath("//Define")
        defines.each do |define|
          define.content = define.content.split(",").
            collect { |d| d.gsub(/USE_BPP=\d+/, "USE_BPP=#{bpp}")}.
            join(",")
        end
      end

      def update_library
        libraries = @project_node.xpath("//Group/Files/File[FileType=4]")
        libraries.each do |l|
          file_path = l.at_xpath("FilePath")
          file_path.content = file_path.content.gsub(/.*touchgfx[\\\/]lib/, "#{path_to(touchgfx_path)}/lib")
        end
      end

      def path_to(file_path)
        absolute_file_path = File.absolute_path(file_path, Dir.pwd).downcase
        project_dir = File.dirname(project_file)
        absolute_project_dir = File.absolute_path(project_dir, Dir.pwd).downcase
        relative_path = Pathname.new(absolute_file_path).relative_path_from(Pathname.new(absolute_project_dir))
        return relative_path.to_s
      end

      def project_file_contents
        File.read(project_file, :encoding=>'utf-8')
      end

      def project_file_xml
        @project_file_xml ||= Nokogiri::XML(project_file_contents) do |config|
          config.default_xml.noblanks
        end
      end

      def output_project_file
        output = project_file_xml.to_xml(:indent => 2,
          :save_with => Nokogiri::XML::Node::SaveOptions::NO_EMPTY_TAGS | Nokogiri::XML::Node::SaveOptions::FORMAT)
        if(output != project_file_contents)
          File.open(project_file, 'wb') { |file| file.write(output) }
        end
      end

      def keil_file_order
        proc { |word| File.basename(word).downcase }
      end

      def abort(msg)
        STDERR.puts msg; Kernel.abort
      end
    end
  end
end
