require 'touchgfx/generators/base'
require 'nokogiri'
require 'securerandom' # guids

module TouchGFX
  module Generators
    class MSVS < Base

      argument :project_file
      argument :bpp, :optional => true
      argument :nonopaque, :optional => true
      argument :bwrle, :optional => true
      PROJ_DIR = '..\..'
      # These are the UUID's for empty application (full and standard version)
      EMPTY_APP_UUID1 = '3C47683C-0505-487F-A1FD-75B8490BF72C'
      EMPTY_APP_UUID2 = 'B87C9F4F-DD87-4881-87CE-01CDABD73E39'

      def check_project_node
        @touchgfx_project = project_file.match("/TouchGFX\.vcxproj$")
        @standard_project = false
        unless @touchgfx_project
          if !project_file.match("/Application.vcxproj$") and !project_file.match("/Application-std.vcxproj$")
            say_status :warning, "You have #{project_file} in EmptyApplication/simulator/msvs/", :red
            say_status :warning, "Only Application.vcxproj and Application-std.vcxproj are supported.", :red
          end
          @standard_project = project_file_xml.xpath('/xmlns:Project/xmlns:ItemGroup/xmlns:ProjectReference/@Include').none? do |f|
            f.to_s.match(/TouchGFX\.vcxproj$/i)
          end
        end
        if project_file_xml.xpath('/xmlns:Project').nil?
          say_status :conflict, "No <project> node in #{project_file}", :red # #{project_file}
          Kernel.abort
        end
      end

      def project_file_contents
        @tgfx_path = "#{ENV['touchgfx'].strip}/touchgfx".gsub('\\','/')

        find_sdl_version

        update_config
        update_sln
        update_vcxproj
        update_vcxproj_filters
        output_project_file
      end

      private

      def uuid(key)
        @existing_uuids ||= {}
        if @existing_uuids[key]
          @existing_uuids[key]
        else
          SecureRandom.uuid.upcase
        end
      end

      def find_sdl_version
        @sdl_version=1
        @subsystem=0 #main
        if File.exist?('simulator/main.cpp')
          contents = File.read('simulator/main.cpp', :encoding=>'utf-8')
          if contents.match(/HALSDL2/)
            @sdl_version=2
          end
          if contents.match(/int CALLBACK WinMain/)
            @subsystem=1 #windows
          end
        end
      end

      def update_config
        if File.exist?(@properties_file)
          touchgfx_release_path = @properties_file_xml.xpath('xmlns:Project/xmlns:PropertyGroup/xmlns:TouchGFXReleasePath')
          unless touchgfx_release_path.empty?
            if tgfx_rel_path.match(/:/)
              touchgfx_release_path[0].content = "#{tgfx_rel_path}"
            else
              #touchgfx_release_path[0].content = "$(ProjectDir)#{tgfx_rel_path}"
              touchgfx_release_path[0].content = "#{tgfx_rel_path}"
            end
          end

          use_bpp = @properties_file_xml.xpath('xmlns:Project/xmlns:PropertyGroup/xmlns:UseBPP')
          new_bpp = bpp == 18 ? 24 : bpp == 0 ? 16 : bpp
          if use_bpp.empty?
            node = Nokogiri::XML::Node.new 'UseBPP', @properties_file_xml
            @properties_file_xml.xpath('xmlns:Project/xmlns:PropertyGroup').first.add_child(node)
            use_bpp = @properties_file_xml.xpath('xmlns:Project/xmlns:PropertyGroup/xmlns:UseBPP')
          else
            if bpp == 0
              new_bpp = use_bpp[0].content.to_i
              if new_bpp == 0
                new_bpp = 16
              end
            end
          end
          use_bpp[0].content = new_bpp

          opaque = @properties_file_xml.xpath('xmlns:Project/xmlns:PropertyGroup/xmlns:OpaqueImageOutputFormat')
          if opaque.empty?
            node = Nokogiri::XML::Node.new 'OpaqueImageOutputFormat', @properties_file_xml
            @properties_file_xml.xpath('xmlns:Project/xmlns:PropertyGroup').first.add_child(node)
            opaque = @properties_file_xml.xpath('xmlns:Project/xmlns:PropertyGroup/xmlns:OpaqueImageOutputFormat')
          end
          case bpp
          when 1
            opaque[0].content = bwrle ? 'BW_RLE' : 'BW'
          when 2
            opaque[0].content = 'GRAY2'
          when 4
            opaque[0].content = 'GRAY4'
          when 16
            opaque[0].content = 'RGB565'
          when 18
            opaque[0].content = 'RGB666'
          when 24
            opaque[0].content = 'RGB888'
          end

          nonopaque = @properties_file_xml.xpath('xmlns:Project/xmlns:PropertyGroup/xmlns:NonOpaqueImageOutputFormat')
          if nonopaque.empty?
            node = Nokogiri::XML::Node.new 'NonOpaqueImageOutputFormat', @properties_file_xml
            @properties_file_xml.xpath('xmlns:Project/xmlns:PropertyGroup').first.add_child(node)
            nonopaque = @properties_file_xml.xpath('xmlns:Project/xmlns:PropertyGroup/xmlns:NonOpaqueImageOutputFormat')
          end
          case bpp
          when 1
            nonopaque[0].content = bwrle ? 'BW_RLE' : 'BW'
          when 2
            nonopaque[0].content = 'GRAY2'
          when 4
            nonopaque[0].content = 'GRAY4'
          when 16
            nonopaque[0].content = nonopaque == 'rgb565' ? 'RGB565' : 'ARGB8888'
          when 18
            nonopaque[0].content = 'ARGB8888'
          when 24
            nonopaque[0].content = 'ARGB8888'
          end


        end
      end

      def update_sln
        vcxproj = File.basename(project_file)

        uuid_match = @solution_content.match("^Project.* = \".*#{vcxproj}.*\"{(.*)}\"")
        unless uuid_match
          say_status :error, "#{File.basename(@solution_file)} does not reference #{vcxproj}, unable to update solution", :red
        else
          old_uuid = uuid_match[1]
          @generate_new_uuids = ((old_uuid == EMPTY_APP_UUID1 || old_uuid == EMPTY_APP_UUID2) && !Dir.pwd.match(/EmptyApplication/))
          if @generate_new_uuids
            @new_uuid = uuid("")
            @solution_content.gsub!(old_uuid, @new_uuid)
          else
            @new_uuid = old_uuid
          end

          tgfx_match = @solution_content.match("^Project.* = \"TouchGFX\", \"(.*\\\\touchgfx)\\\\.*\", \".*\"$")
          if tgfx_match
            old_path = tgfx_match[1]
            @solution_content.gsub!("#{old_path}", "#{tgfx_rel_path}")
          end
        end
      end

      def remove_unless_exist(node)
        file_name = node['Include']
        if file_name.start_with?(PROJ_DIR)
          file_name = file_name[PROJ_DIR.length+1 .. -1]
        elsif file_name.start_with?('$(TouchGFXReleasePath)')
          file_name = "#{@tgfx_path}#{file_name[22..-1]}"
        end
        if !File.exist?(file_name)
          node.remove
        elsif !@touchgfx_project
          if @sdl_version == 1
            node.remove if file_name.match(/\/sdl2\//) || file_name.match(/\\sdl2\\/) || file_name.match(/SDL2Touch/)
          elsif @sdl_version == 2
            node.remove if file_name.match(/\/sdl\//) || file_name.match(/\\sdl\\/) || file_name.match(/SDLTouch/)
          end
        end
      end

      def add_node_unless_exist(type, file_name)
        parent = @project_item_group_clcompile if type=='ClCompile'
        parent = @project_item_group_clinclude if type=='ClInclude'
        parent = @project_item_group_none unless parent
        if parent.xpath("xmlns:#{type}[@Include='#{file_name}']").empty?
          node = Nokogiri::XML::Node.new type, @project_file_xml
          node['Include'] = file_name
          parent.add_child(node)
          return node
        else
          return nil
        end
      end

      def update_vcxproj
        project_node = @project_file_xml.xpath('/xmlns:Project')

        # Update SDL version
        import = project_node.xpath("xmlns:ImportGroup[@Label='PropertySheets']/xmlns:Import[starts-with(@Project,'$(TouchGFXReleasePath)\\config\\msvs\\touchgfx')]")
        if @sdl_version == 1
          import.each { |n| n['Project'] = '$(TouchGFXReleasePath)\config\msvs\touchgfx.props' }
        elsif @sdl_version == 2
          import.each { |n| n['Project'] = '$(TouchGFXReleasePath)\config\msvs\touchgfx_sdl2.props' }
        end

        # Update copy commands
        cmd = project_node.xpath('xmlns:ItemDefinitionGroup/xmlns:PostBuildEvent/xmlns:Command')
        if @sdl_version == 1
          cmd.each { |n| n.content = "copy $(TouchGFXReleasePath)\\lib\\sdl\\win32\\SDL.dll $(OutDir)\ncopy $(TouchGFXReleasePath)\\lib\\sdl\\win32\\SDL_image.dll $(OutDir)\n" }
        elsif @sdl_version == 2
          cmd.each { |n| n.content = "copy $(TouchGFXReleasePath)\\lib\\sdl2\\win32\\SDL2.dll $(OutDir)\ncopy $(TouchGFXReleasePath)\\lib\\sdl2\\win32\\SDL2_image.dll $(OutDir)\ncopy $(TouchGFXReleasePath)\\lib\\sdl2\\win32\\libpng16-16.dll $(OutDir)\ncopy $(TouchGFXReleasePath)\\lib\\sdl2\\win32\\zlib1.dll $(OutDir)\nif exist ..\\landscape.png copy /y ..\\landscape.png $(OutDir)\nif exist ..\\portrait.png copy /y ..\\portrait.png $(OutDir)\n" }
        end

        # Update SubSystem
        subsys = project_node.xpath('xmlns:ItemDefinitionGroup/xmlns:Link/xmlns:SubSystem')
        subsys.each do |n|
          n.content = @subsystem == 0 ? 'Console' : 'Windows'
        end

        @project_item_group_clcompile = project_node.xpath('xmlns:ItemGroup/xmlns:ClCompile/..')
        if @project_item_group_clcompile.empty?
          @project_item_group_clcompile = Nokogiri::XML::Node.new 'ItemGroup', @project_file_xml
          # project_node.children.first is "ProjectConfigutations"
          project_node.children.first.next = @project_item_group_clcompile
        else
          @project_item_group_clcompile = @project_item_group_clcompile[0]
        end

        @project_item_group_clinclude = project_node.xpath('xmlns:ItemGroup/xmlns:ClInclude/..')
        if @project_item_group_clinclude.empty?
          @project_item_group_clinclude = Nokogiri::XML::Node.new 'ItemGroup', @project_file_xml
          @project_item_group_clcompile.next = @project_item_group_clinclude
        else
          @project_item_group_clinclude = @project_item_group_clinclude[0]
        end

        @project_item_group_none = project_node.xpath('xmlns:ItemGroup/xmlns:None/..')
        if @project_item_group_none.empty?
          @project_item_group_none = Nokogiri::XML::Node.new 'ItemGroup', @project_file_xml
          @project_item_group_clinclude.add_next_sibling @project_item_group_none
        else
          @project_item_group_none = @project_item_group_none[0]
        end

        project_node.at_xpath('xmlns:PropertyGroup/xmlns:ProjectGuid').content = "{#{@new_uuid}}"
        # Do not set RootNamespace and ProjectName at the moment
        #name = File.basename(Dir.pwd)
        #project_node.at_xpath('xmlns:PropertyGroup/xmlns:RootNamespace').content = name
        #project_node.at_xpath('xmlns:PropertyGroup/xmlns:ProjectName').content = name

        # Clean out old 'ClCompile', 'ClInclude' & 'None' for non-existing files
        @project_item_group_clcompile.children.each{|node| remove_unless_exist(node)}
        @project_item_group_clinclude.children.each{|node| remove_unless_exist(node)}
        @project_item_group_none.children.each{|node| remove_unless_exist(node)}

        if @touchgfx_project
          @cpp_files_touchgfx = []
          @cpp_files = Dir["touchgfx/framework/source/common/**/*.cpp",
                           "touchgfx/framework/source/platform/driver/lcd/**/*.cpp",
                           "touchgfx/framework/source/platform/driver/touch/SDLTouch*.cpp", # Avoid SDL2Touch*
                           "touchgfx/framework/source/platform/hal/simulator/sdl/*.cpp", # Avoid simulator/sdl2/*.cpp
                           "touchgfx/framework/source/touchgfx/**/*.cpp"]
          @cpp_files.sort_by(&msvs_file_order).each do |file_name|
            node = Nokogiri::XML::Node.new 'ClCompile', @project_file_xml
            node['Include'] = "../../../../#{file_name}".gsub('/../touchgfx/','/../').gsub('/','\\')
            @project_item_group_clcompile.add_child node
          end

          @hpp_files_touchgfx = []
          @hpp_files = Dir["touchgfx/framework/include/common/**/*.hpp",
                           "touchgfx/framework/include/mvp/**/*.hpp",
                           "touchgfx/framework/include/platform/driver/lcd/**/*.hpp",
                           "touchgfx/framework/include/platform/hal/simulator/sdl/*.hpp",
                           "touchgfx/framework/include/touchgfx/**/*.hpp"]
          @hpp_files.sort_by(&msvs_file_order).each do |file_name|
            node = Nokogiri::XML::Node.new 'ClInclude', @project_item_group_clinclude
            node['Include'] = "../../../../#{file_name}".gsub('/../touchgfx/','/../').gsub('/','\\')
            @project_item_group_clinclude.add_child node
          end

          @resource_files = ['changelog.txt','known_issues.txt']
          @resource_files.each do |file_name|
            node = Nokogiri::XML::Node.new 'None', @project_item_group_none
            node['Include'] = "../../../../../#{file_name}".gsub('/','\\')
            @project_item_group_none.add_child node
          end
        else
          @cpp_files = Dir['gui/src/**/*.cpp', 'gui/src/**/*.c',
                           'simulator/src/**/*.cpp', 'simulator/src/**/*.c',
                           'simulator/main.cpp',
                           'generated/gui_generated/src/**/*.cpp'];
          if @sdl_version == 1
            @cpp_files_touchgfx = Dir["#{@tgfx_path}/framework/source/platform/driver/touch/SDLTouchController.cpp",
                                      "#{@tgfx_path}/framework/source/platform/hal/simulator/sdl/*.cpp"]
          elsif @sdl_version == 2
            @cpp_files_touchgfx = Dir["#{@tgfx_path}/framework/source/platform/driver/touch/SDL2TouchController.cpp",
                                      "#{@tgfx_path}/framework/source/platform/hal/simulator/sdl2/*.cpp"]
          end
          (@cpp_files+@cpp_files_touchgfx).sort_by(&msvs_file_order).each do |file_name|
            add_node_unless_exist('ClCompile', "../../#{file_name}".gsub("../../#{@tgfx_path}",'$(TouchGFXReleasePath)').gsub('/','\\'))
          end

          @hpp_files = Dir['gui/include/**/*.hpp',
                           'gui/include/**/*.h',
                           'simulator/include/**/*.hpp',
                           'simulator/include/**/*.h',
                           'generated/gui_generated/include/**/*.hpp']
          @hpp_files_touchgfx = []
          if @standard_project
            @hpp_files_touchgfx = Dir["#{@tgfx_path}/framework/include/common/**/*.hpp",
                                      "#{@tgfx_path}/framework/include/mvp/**/*.hpp",
                                      "#{@tgfx_path}/framework/include/platform/driver/lcd/**/*.hpp",
                                      "#{@tgfx_path}/framework/include/touchgfx/**/*.hpp"]
          end
          if @sdl_version == 1
            @hpp_files_touchgfx += Dir["#{@tgfx_path}/framework/include/platform/driver/touch/SDLTouchController.hpp",
                                       "#{@tgfx_path}/framework/include/platform/hal/simulator/sdl/*.hpp"]
          elsif @sdl_version == 2
            @hpp_files_touchgfx += Dir["#{@tgfx_path}/framework/include/platform/driver/touch/SDL2TouchController.hpp",
                                       "#{@tgfx_path}/framework/include/platform/hal/simulator/sdl2/*.hpp"]
          end

          (@hpp_files+@hpp_files_touchgfx).sort_by(&msvs_file_order).each do |file_name|
            add_node_unless_exist('ClInclude', "../../#{file_name}".gsub("../../#{@tgfx_path}",'$(TouchGFXReleasePath)').gsub('/','\\'))
          end

          # Remember to use "uniq" as e.g. .bmp and .BMP will find the
          # same file twice on Windows.
          @resource_files = Dir['assets/images/**/*.bmp', 'assets/images/**/*.BMP',
                                'assets/images/**/*.png', 'assets/images/**/*.PNG',
                                'assets/texts/**/*.xlsx', 'assets/fonts/**/*.ttf',
                                'assets/texts/**/*.otf', 'assets/fonts/**/*.bdf',
                                'config/msvs/*'].uniq.sort_by(&msvs_file_order)
          @resource_files.reject!{|f|f.match('~\$texts.xlsx$')}
          @resource_files.each do |file_name|
            add_node_unless_exist('None', "..\\..\\#{file_name.gsub('/','\\')}")
          end
        end

        @project_item_group_clcompile.remove if @project_item_group_clcompile and @project_item_group_clcompile.children.empty?
        @project_item_group_clinclude.remove if @project_item_group_clinclude and @project_item_group_clinclude.children.empty?
        @project_item_group_none.remove if @project_item_group_none and @project_item_group_none.children.empty?

        # Sort children
        @project_item_group_clcompile.children.remove.sort_by(&include_sort).each do |node|
          include = node['Include']
          @project_item_group_clcompile.add_child(node) if @project_item_group_clcompile.xpath("xmlns:ClCompile[@Include='#{include}']").empty?
        end
        @project_item_group_clinclude.children.remove.sort_by(&include_sort).each do |node|
          include = node['Include']
          @project_item_group_clinclude.add_child(node) if @project_item_group_clinclude.xpath("xmlns:ClInclude[@Include='#{include}']").empty?
        end
        @project_item_group_none.children.remove.sort_by(&include_sort).each do |node|
          include = node['Include']
          @project_item_group_none.add_child(node) if @project_item_group_none.xpath("xmlns:None[@Include='#{include}']").empty?
        end
      end

      def update_vcxproj_filters
        project_filter_node = @project_filter_file_xml.xpath('/xmlns:Project')

        @filters_item_group_filter = project_filter_node.xpath('xmlns:ItemGroup/xmlns:Filter/..')
        if @filters_item_group_filter.empty?
          @filters_item_group_filter = Nokogiri::XML::Node.new 'ItemGroup', @project_filter_file_xml
          project_filter_node.children.first.before(@filters_item_group_filter)
        else
          @filters_item_group_filter = @filters_item_group_filter[0]
        end

        @filters_item_group_clcompile = project_filter_node.xpath('xmlns:ItemGroup/xmlns:ClCompile/..')
        if !@filters_item_group_clcompile
          @filters_item_group_clcompile = Nokogiri::XML::Node.new 'ItemGroup', @project_filter_file_xml
          @filters_item_group_filter.next = @filters_item_group_clcompile
        else
          @filters_item_group_clcompile = @filters_item_group_clcompile[0]
        end

        @filters_item_group_clinclude = project_filter_node.xpath('xmlns:ItemGroup/xmlns:ClInclude/..')
        if !@filters_item_group_clinclude
          @filters_item_group_clinclude = Nokogiri::XML::Node.new 'ItemGroup', @project_filter_file_xml
          @filters_item_group_clcompile.next = @filters_item_group_clinclude
        else
          @filters_item_group_clinclude = @filters_item_group_clinclude[0]
        end

        @filters_item_group_none = project_filter_node.xpath('xmlns:ItemGroup/xmlns:None/..')
        if !@filters_item_group_none
          @filters_item_group_none = Nokogiri::XML::Node.new 'ItemGroup', @project_filter_file_xml
          @filters_item_group_clinclude.next = @filters_item_group_none
        else
          @filters_item_group_none = @filters_item_group_none[0]
        end

        @existing_uuids ||= {}
        if !@generate_new_uuids
          # Save existing uuids!
          @filters_item_group_filter.children.each do |node|
            @existing_uuids[ node['Include'] ] = node.xpath('xmlns:UniqueIdentifier').text.gsub('{','').gsub('}','')
          end
        end

        # Clean out old 'ClCompile', 'ClInclude' & 'None'
        @filters_item_group_clcompile.children.each{|node| remove_unless_exist(node)}
        @filters_item_group_clinclude.children.each{|node| remove_unless_exist(node)}
        @filters_item_group_none.children.each{|node| remove_unless_exist(node)}

        # Find all nodes starting with a specific string
        ['Header Files\\gui','Header Files\\gui_generated','Header Files\\TouchGFX',
         'Source Files\\gui','Source Files\\gui_generated','Source Files\\TouchGFX',
         'Resource Files\\assets','Resource Files\\gui'].each do |filter|
          @filters_item_group_filter.xpath("xmlns:Filter[@Include='#{filter}']").each do |node|
            node.remove
          end
          @filters_item_group_filter.xpath("xmlns:Filter[starts-with(@Include,'#{filter}\\')]").each do |node|
            node.remove
          end
        end

        if @filters_item_group_filter.xpath("xmlns:Filter[@Include='Source Files']").empty?
          node = Nokogiri::XML::Node.new 'Filter', @filters_item_group_filter
          node['Include'] = 'Source Files'
          unique_uuid = Nokogiri::XML::Node.new 'UniqueIdentifier', node
          unique_uuid.content = "{#{uuid(node['Include'])}}"
          extensions =  Nokogiri::XML::Node.new "Extensions", node
          extensions.content = 'cpp;c;cc;cxx;def;odl;idl;hpj;bat;asm;asmx'
          node.add_child unique_uuid
          node.add_child extensions
          @filters_item_group_filter.add_child node
        end

        if @filters_item_group_filter.xpath("xmlns:Filter[@Include='Header Files']").empty?
          node = Nokogiri::XML::Node.new 'Filter', @filters_item_group_filter
          node['Include'] = 'Header Files'
          unique_uuid = Nokogiri::XML::Node.new 'UniqueIdentifier', node
          unique_uuid.content = "{#{uuid(node['Include'])}}"
          extensions = Nokogiri::XML::Node.new 'Extensions', node
          extensions.content = 'h;hpp;hxx;hm;inl;inc;xsd'
          node.add_child unique_uuid
          node.add_child extensions
          @filters_item_group_filter.add_child node
        end

        if @filters_item_group_filter.xpath("xmlns:Filter[@Include='Resource Files']").empty?
          node = Nokogiri::XML::Node.new 'Filter', @filters_item_group_filter
          node['Include'] = 'Resource Files'
          unique_uuid = Nokogiri::XML::Node.new 'UniqueIdentifier', node
          unique_uuid.content = "{#{uuid(node['Include'])}}"
          extensions = Nokogiri::XML::Node.new 'Extensions', node
          extensions.content = 'rc;ico;cur;bmp;dlg;rc2;rct;bin;rgs;gif;jpg;jpeg;jpe;resx;tiff;tif;ttf;otf;bdf;png;wav;xlsx;props;mfcribbon-ms'
          node.add_child unique_uuid
          node.add_child extensions
          @filters_item_group_filter.add_child node
        end

        @hpp_files.sort_by(&msvs_file_order).each do |file_name|
          dir_name = File.dirname(file_name)
          folder = dir_name.match("^gui/include/gui(.*)$")
          folder_name = "Header Files/gui#{folder[1]}" if folder
          unless folder
            folder = dir_name.match("^gui(.*)$")
            folder_name = "Header Files/gui#{folder[1]}" if folder
          end
#          unless folder
#            folder = dir_name.match("^(simulator)/include(.*)$")
#            folder_name = "Header Files/gui/#{folder[1-2]}" if folder
#          end
          unless folder
            folder = dir_name.match("^touchgfx/framework/include(.*)$")
            if @touchgfx_project
              folder_name = "Header Files#{folder[1]}" if folder
            else
              folder_name = "Header Files/TouchGFX#{folder[1]}" if folder
            end
          end
          unless folder
            folder = dir_name.match("^generated/gui_generated/include/gui_generated(.*)$")
            folder_name = "Header Files/gui_generated#{folder[1]}" if folder
          end
          unless folder
            folder = ('/'+dir_name).match("^(.*)$")
            folder_name = "Header Files/gui#{folder[1]}" if folder
          end
          unless folder
            fail
          end

          add_filter folder_name

          # Create node for <ClInclude>
          node = Nokogiri::XML::Node.new 'ClInclude', @project_filter_file_xml
          include = ''
          if @touchgfx_project
            include = "../../../../#{file_name}".gsub('/../touchgfx/','/../').gsub('/','\\')
          else
            include = "../../#{file_name}".gsub("../../#{@tgfx_path}",'$(TouchGFXReleasePath)').gsub('/','\\')
          end
          node['Include'] = include
          filter = Nokogiri::XML::Node.new 'Filter', node
          filter.content = folder_name.gsub('/','\\')
          node.add_child filter
          hpp_node = @filters_item_group_clinclude.xpath("xmlns:ClInclude[@Include='#{include}']")
          if hpp_node.empty?
            @filters_item_group_clinclude.add_child(node)
          else
            hpp_node.each{|node|node.xpath("xmlns:Filter")[0].content = filter.content}
            node.remove
          end
        end

        @hpp_files_touchgfx.sort_by(&msvs_file_order).each do |file_name|
          folder_name = File.dirname(file_name).gsub("#{@tgfx_path}/framework/include/",'Header Files/TouchGFX/')
          add_filter folder_name
          # Create node for <ClInclude>
          node = Nokogiri::XML::Node.new "ClInclude", @project_filter_file_xml
          include = "../../#{file_name}".gsub("../../#{@tgfx_path}",'$(TouchGFXReleasePath)').gsub('/','\\')
          node['Include'] = include
          filter = Nokogiri::XML::Node.new 'Filter', node
          filter.content = folder_name.gsub('/','\\')
          node.add_child filter
          hpp_node = @filters_item_group_clinclude.xpath("xmlns:ClInclude[@Include='#{include}']")
          if hpp_node.empty?
            @filters_item_group_clinclude.add_child(node)
          else
            hpp_node.each{|node|node.xpath("xmlns:Filter")[0].content = filter.content}
            node.remove
          end
        end

        @cpp_files.sort_by(&msvs_file_order).each do |file_name|
          dir_name = File.dirname(file_name)
          folder = dir_name.match("^gui/src(.*)$")
          folder_name = "Source Files/gui#{folder[1]}" if folder
          unless folder
            folder = dir_name.match("^touchgfx/framework/source(.*)$")
            if @touchgfx_project
              folder_name = "Source Files#{folder[1]}" if folder
            else
              folder_name = "Source Files/TouchGFX#{folder[1]}" if folder
            end
          end
          unless folder
            folder = dir_name.match("generated(/gui_generated)/src(.*)$")
            folder_name = "Source Files#{folder[1]}#{folder[2]}" if folder
          end
          unless folder
            folder = ('/'+dir_name).match("^(.*)$")
            folder_name = "Source Files/gui#{folder[1]}" if folder
          end
          unless folder
            fail
          end

          add_filter folder_name

          # Create node for <ClInclude>
          node = Nokogiri::XML::Node.new "ClCompile", @project_filter_file_xml
          include=''
          if @touchgfx_project
            include = "../../../../#{file_name}".gsub('/../touchgfx/','/../').gsub('/','\\')
          else
            include = "../../#{file_name}".gsub("../../#{@tgfx_path}",'$(TouchGFXReleasePath)').gsub('/','\\')
          end
          node['Include'] = include
          filter = Nokogiri::XML::Node.new 'Filter', node
          filter.content = folder_name.gsub('/','\\')
          node.add_child filter
          cpp_node = @filters_item_group_clcompile.xpath("xmlns:ClCompile[@Include=\'#{include}\']")
          if cpp_node.empty?
            @filters_item_group_clcompile.add_child(node)
          else
            cpp_node.each{|node|node.xpath("xmlns:Filter")[0].content = filter.content}
            node.remove
          end
        end

        @cpp_files_touchgfx.sort_by(&msvs_file_order).each do |file_name|
          folder_name = File.dirname(file_name).gsub("#{@tgfx_path}/framework/source/",'Source Files/TouchGFX/')
          add_filter folder_name
          # Create node for <ClInclude>
          node = Nokogiri::XML::Node.new "ClCompile", @project_filter_file_xml
          include = "../../#{file_name}".gsub("../../#{@tgfx_path}",'$(TouchGFXReleasePath)').gsub('/','\\')
          node['Include'] = include
          filter = Nokogiri::XML::Node.new 'Filter', node
          filter.content = folder_name.gsub('/','\\')
          node.add_child filter
          cpp_node = @filters_item_group_clcompile.xpath("xmlns:ClCompile[@Include=\'#{include}\']")
          if cpp_node.empty?
            @filters_item_group_clcompile.add_child(node)
          else
            cpp_node.each{|node|node.xpath("xmlns:Filter")[0].content = filter.content}
            node.remove
          end
        end

        @resource_files.each do |file_name|
          folder = file_name.match("^assets/(.*)/")
          folder_name = "Resource Files"
          if folder
            folder_name = "#{folder_name}/assets/#{folder[1..-1].to_a.join}"
          else
            folder = file_name.match("^config/(.*)/.*")
            if folder
              folder_name = "#{folder_name}/config/#{folder[1..-1].to_a.join}"
            end
          end
          add_filter folder_name

          # Create node for <ClInclude>
          node = Nokogiri::XML::Node.new "None", @project_filter_file_xml
          include = ''
          if @touchgfx_project
            include = "../../../../../#{file_name}".gsub('/','\\')
          else
            include = "../../#{file_name}".gsub('/','\\')
          end
          node['Include'] = include
          filter = Nokogiri::XML::Node.new 'Filter', node
          filter.content = folder_name.gsub('/','\\')
          node.add_child filter
          res_node = @filters_item_group_none.xpath("xmlns:None[@Include=\'#{include}\']")
          if res_node.empty?
            @filters_item_group_none.add_child(node)
          else
            res_node.each{|node|node.xpath("xmlns:Filter")[0].content = filter.content}
            node.remove
          end
        end

        # Remove empty filters
        @filters_item_group_filter.children.each do |node|
          path = node['Include']
          if @project_filter_file_xml.xpath("//xmlns:Filter[text()='#{path}']").empty? &&
             @project_filter_file_xml.xpath("//xmlns:Filter[starts-with(text(),'#{path}\\')]").empty?
            node.remove
          end
        end

        @filters_item_group_filter.remove if @filters_item_group_filter and @filters_item_group_filter.children.empty?
        @filters_item_group_clcompile.remove if @filters_item_group_clcompile and @filters_item_group_clcompile.children.empty?
        @filters_item_group_clinclude.remove if @filters_item_group_clinclude and @filters_item_group_clinclude.children.empty?
        @filters_item_group_none.remove if @filters_item_group_none and @filters_item_group_none.children.empty?

        # Sort children
        @filters_item_group_filter.children.remove.sort_by(&include_node_breadth_first).each do |node|
          include = node['Include']
          @filters_item_group_filter.add_child(node) if @filters_item_group_filter.xpath("xmlns:ClCompile[@Include='#{include}']").empty?
        end
        @filters_item_group_clcompile.children.remove.sort_by(&include_sort).each do |node|
          include = node['Include']
          @filters_item_group_clcompile.add_child(node) if @filters_item_group_clcompile.xpath("xmlns:ClCompile[@Include='#{include}']").empty?
        end
        @filters_item_group_clinclude.children.remove.sort_by(&include_sort).each do |node|
          include = node['Include']
          @filters_item_group_clinclude.add_child(node) if @filters_item_group_clinclude.xpath("xmlns:ClInclude[@Include='#{include}']").empty?
        end
        @filters_item_group_none.children.remove.sort_by(&include_sort).each do |node|
          include = node['Include']
          @filters_item_group_none.add_child(node) if @filters_item_group_none.xpath("xmlns:None[@Include='#{include}']").empty?
        end
      end

      def add_filter(name)
        dosname = name.gsub("/", "\\")
        dosname_quoted = dosname.gsub("\\", "\\\\")

        if @filters_item_group_filter.xpath("xmlns:Filter[@Include='#{dosname_quoted}']").empty?
          if name.match("/")
            add_filter(File.dirname(name))
          end
        end

        if @filters_item_group_filter.xpath("xmlns:Filter[@Include='#{dosname_quoted}']").empty?
          # Create node for <Filter>
          node = Nokogiri::XML::Node.new 'Filter', @project_filter_file_xml
          node['Include'] = dosname
          unique_uuid = Nokogiri::XML::Node.new "UniqueIdentifier", node
          unique_uuid.content = "{#{uuid(dosname)}}"
          node.add_child unique_uuid
          @filters_item_group_filter.add_child(node)
        end
      end

      def group_with_name(parent, name)
        element_with_type_name_and_contents(parent, 'group', name, name)
      end

      def file_with_file_name(parent, file_name)
        path = "$PROJ_DIR$\\..\\..\\..\\..\\#{file_name.gsub('/', '\\')}".
          gsub("\\..\\target", "") #special case for the target folder, could use Pathname#cleanpath instead
        element_with_type_name_and_contents(parent, 'file', file_name, path)
      end

      def element_with_type_name_and_contents(parent, type, name, contents)
        parent.at_xpath("./#{type}[name=\"#{contents}\"]") ||
          begin
            #say_status :create, "#{type} #{name}", :green
            node = Nokogiri::XML::Node.new type, project_file_xml
            name_node = Nokogiri::XML::Node.new 'name', project_file_xml
            name_node.content = contents
            name_node.parent = node
            parent.add_child node
          end
      end

      ### ALSO IN IAR.RB ###
      def nice_path(path)
        path.gsub!('/','\\')
        path = path[0..-2] if path[-1]=='\\'
        return path
      end

      def tgfx_rel_path
        unless @tgfx_rel_path
          tgfx_array = @tgfx_path.gsub(/^(.:)/){$1.downcase}.gsub('\\','/').split('/')
          cur_array = Dir.pwd.gsub(/^(.:)/){$1.downcase}.gsub('\\','/').split('/')
          while tgfx_array[0] == cur_array[0]
            tgfx_array = tgfx_array[1..-1]
            cur_array = cur_array[1..-1]
          end
          @tgfx_rel_path = File.join(['..']*cur_array.size, tgfx_array)
          if PROJ_DIR.length > 0
            @tgfx_rel_path = File.join(PROJ_DIR, @tgfx_rel_path)
          end
          if @tgfx_rel_path.match(/:/)
            @tgfx_rel_path = @tgfx_path
          elsif @tgfx_path.start_with?('..')
            @tgfx_rel_path = File.join(PROJ_DIR, @tgfx_path)
          end
          @tgfx_rel_path = nice_path(@tgfx_rel_path)
        end
        return @tgfx_rel_path
      end

      def relative_path(file_path)
        cur_path = Dir.pwd
        proj_path = PROJ_DIR

        if file_path.start_with?(File.join(@tgfx_path,'touchgfx').gsub('/','\\'))
          return nice_path(File.join(tgfx_rel_path, file_path[@tgfx_path.length..-1]))
        elsif file_path.start_with?(cur_path)
          return nice_path(File.join(proj_path, file_path[cur_path.length..-1]))
        end
        return nice_path(File.join(proj_path, file_path))
      end

      def write_file(file, output, force=false)
        input = File.read(file, :encoding=>'utf-8')
        if force || input != output
          File.open(file, 'w:utf-8') { |f| f.write(output) }
          return true
        end
        return false
      end

      def target
        File.join(Dir.pwd)
      end

      def project_file_xml
        @properties_file = "config/msvs/Application.props"
        if File.exist?(@properties_file)
          @properties_file_xml = Nokogiri::XML(File.read(@properties_file, :encoding=>'utf-8')) do |config|
            config.default_xml.noblanks
          end
        end

        @solution_file = "#{project_file.gsub('vcxproj','sln')}"
        @solution_content = File.read(@solution_file, :encoding => 'utf-8')

        @project_filter_file = "#{project_file}.filters"
        @project_filter_file_xml ||= Nokogiri::XML(File.read(@project_filter_file, :encoding=>'utf-8')) do |config|
          config.default_xml.noblanks
        end

        @project_file_xml ||= Nokogiri::XML(File.read(project_file, :encoding=>'utf-8')) do |config|
          config.default_xml.noblanks
        end
      end

      def output_project_file
        if File.exist?(@properties_file)
          write_file(@properties_file, @properties_file_xml.to_xml(:indent => 2).gsub("<TextDataFormat/>","<TextDataFormat></TextDataFormat>").gsub("/>", " />"))
        end
        wrote_filter = write_file(@project_filter_file, @project_filter_file_xml.to_xml(:indent => 2).gsub("/>", " />"))
        write_file(project_file, @project_file_xml.to_xml(:indent => 2).gsub("/>", " />").gsub("\n</Project>\n","\n</Project>"), wrote_filter)
        write_file(@solution_file, @solution_content)
      end

      def msvs_file_order
        proc { |word| word.downcase }
      end

      def include_sort
        proc { |node| node['Include'].downcase }
      end

      def include_node_breadth_first
        proc { |node| node['Include'].split('\\').insert(1,node['Include'].count('\\').to_s).join('\\').downcase }
      end
    end
  end
end
