require 'touchgfx/generators/base'
require 'nokogiri'
require 'pathname'
require 'securerandom' # guids

module TouchGFX
  module Generators
    class MSVSProjectUpdater
      attr_reader :project_file
      attr_reader :touchgfx_path
      attr_reader :touchgfx_env
      attr_reader :config 

      def initialize(project_file, touchgfx_path, touchgfx_env, config, gui_group_name = 'gui', generated_group_name = 'generated')
        @project_file = project_file
        @touchgfx_path = touchgfx_path
        @touchgfx_env = touchgfx_env
        @config = config
        @gui_group_name = gui_group_name
        @generated_group_name = generated_group_name
      end

      def run
        check_project
        update_config
        ensure_source_nodes
        output_project_files
      end

      private

      def check_project
        @project_node = ( project_file_xml.at_xpath('.//xmlns:Project') || abort("no <Project> node in #{project_file}") )
        @filters_node = ( filters_file_xml.at_xpath('.//xmlns:Project') || abort("no <Project> node in #{filters_file}") )
        @filters_node.at_xpath("xmlns:ItemGroup/xmlns:Filter[@Include='Source Files']") || abort("no Source Files Filter")
        @filters_node.at_xpath("xmlns:ItemGroup/xmlns:Filter[@Include='Header Files']") || abort("no Header Files Filter")
      end

      def update_config
        abort "no #{config_file} found" if !File.exist?(config_file)
        application_root = config_file_xml.at_xpath('xmlns:Project/xmlns:PropertyGroup/xmlns:ApplicationRoot').content
        config_file_xml.at_xpath('xmlns:Project/xmlns:PropertyGroup/xmlns:TouchGFXReleasePath').content = File.join(application_root, touchgfx_path).gsub('/','\\')
        config_file_xml.at_xpath('xmlns:Project/xmlns:PropertyGroup/xmlns:UseBPP').content = config.bpp
        config_file_xml.at_xpath('xmlns:Project/xmlns:PropertyGroup/xmlns:OpaqueImageOutputFormat').content = config.opaque
        config_file_xml.at_xpath('xmlns:Project/xmlns:PropertyGroup/xmlns:NonOpaqueImageOutputFormat').content = config.nonopaque
        env_path = config_file_xml.at_xpath('xmlns:Project/xmlns:PropertyGroup/xmlns:TouchGFXEnvPath') || begin
          node = Nokogiri::XML::Node.new 'TouchGFXEnvPath', config_file_xml
          parent = config_file_xml.at_xpath('xmlns:Project/xmlns:PropertyGroup/xmlns:TouchGFXReleasePath').parent
          parent.add_child node
        end
        env_path.content = touchgfx_env.gsub('/','\\')
      end

      def ensure_source_nodes
        clcompile_project = @project_node.at_xpath('xmlns:ItemGroup/xmlns:ClCompile/..')
        clinclude_project = @project_node.at_xpath('xmlns:ItemGroup/xmlns:ClInclude/..')
        clcompile_filters = @filters_node.at_xpath('xmlns:ItemGroup/xmlns:ClCompile/..')
        clinclude_filters = @filters_node.at_xpath('xmlns:ItemGroup/xmlns:ClInclude/..')
        clfilters_filters = @filters_node.at_xpath('xmlns:ItemGroup/xmlns:Filter/..')

        # remove non-existing gui or generated files from project
        children = [clcompile_project, clinclude_project, clcompile_filters, clinclude_filters].collect { |n| n.children }.flatten
        path_to_generated = path_to('generated')
        path_to_gui = path_to('gui')
        children.select do |c|
          file_name = c["Include"]
          gui_or_generated_file = file_name.start_with?(path_to_generated) || file_name.start_with?(path_to_gui)
          gui_or_generated_file && !File.exist?( File.join(project_dir, file_name) )
        end.each { |r| r.remove }

        # remove non-existing gui or generated files from filters
        clfilters_filters.children.select do |filter|
          filter_group_name = filter["Include"]
          should_delete = false
          if filter_group_name.start_with?("Header Files\\gui\\") ||
             filter_group_name.start_with?("Header Files\\generated\\") ||
             filter_group_name.start_with?("Source Files\\gui\\") ||
             filter_group_name.start_with?("Source Files\\generated\\")
              should_delete = !Dir.exist?(folder_name(filter_group_name))
          end
        end.each {|node| node.remove}

        # introduce gui and generated files in project and filters
        should_create_compile_or_include = lambda do |group, file_name|
          group.children.none? { |c| c['Include'].downcase == path_to(file_name).downcase } #todo use a hash for speedup #todo never present if we always delete
        end
        create_project_node = lambda do |group_name, file_name|
          Nokogiri::XML::Node.new( group_name, @project_file_xml ).tap { |n| n['Include'] = path_to(file_name) }
        end
        create_filters_node = lambda do |group_name, file_name|
          node = Nokogiri::XML::Node.new( group_name, @filters_file_xml ).tap { |n| n['Include'] = path_to(file_name) }
          filter = Nokogiri::XML::Node.new( 'Filter', node ).tap { |f| f.content = filter_group(file_name) }
          node.add_child filter
          node
        end
        [
          ['{gui,generated/gui_generated}/**/*.{cpp,c}', clcompile_project, 'ClCompile', should_create_compile_or_include, create_project_node],
          ['{gui,generated/gui_generated}/**/*.{hpp,h}', clinclude_project, 'ClInclude', should_create_compile_or_include, create_project_node],
          ['{gui,generated/gui_generated}/**/*.{cpp,h}', clcompile_filters, 'ClCompile', should_create_compile_or_include, create_filters_node],
          ['{gui,generated/gui_generated}/**/*.{hpp,h}', clinclude_filters, 'ClInclude', should_create_compile_or_include, create_filters_node],
        ].each do |pattern, group, group_name, should_create, create|
          Dir.glob(pattern).sort_by {|f| File.basename(f).downcase }.each do |file_name|
            if should_create[group, file_name]
              node = create[group_name, file_name]
              group.add_child node
            end
          end
        end

        #introduce filter groups
        @filters ||= clfilters_filters.children.inject({}) { |m,node| m.merge( {node['Include'] => node} ) } #hash of all existing filters
        Dir.glob('{gui,generated/gui_generated}/**/*.{hpp,cpp,h,c}').collect do |file_name|
          filter_groups(file_name)
        end.flatten.sort_by { |f| f.scan('\\').count }.uniq.each do |group_name|
          if(!@filters.has_key?(group_name))
            node = Nokogiri::XML::Node.new( "Filter", @filters_file_xml ).tap { |n| n['Include'] = group_name }
            filter = Nokogiri::XML::Node.new( 'UniqueIdentifier', node ).tap { |f| f.content = SecureRandom.uuid.upcase }
            node.add_child filter
            clfilters_filters.add_child node
            @filters[group_name] = node
          end
        end
      end

      def folder_name(filter_group)
        # convert gui/generated filter group name to folder name
        type, gui_or_generated, *split_path = filter_group.split('\\')
        type = type == "Header Files" ? true : false
        if gui_or_generated == 'gui'
          directory_path = [ gui_or_generated, type ? ['include', gui_or_generated, split_path] : ['src', split_path]]
        elsif gui_or_generated == 'generated'
          directory_path = [ gui_or_generated, 'gui_generated', type ? ['include', split_path] : ['src', split_path[1..-1]]]
        end
        File.join(directory_path)
      end

      def absolute_project_dir()
        @absolute_project_dir ||= Pathname.new(File.absolute_path(project_dir, Dir.pwd))
      end

      def path_to(file_path)
        absolute_file_path = File.absolute_path(file_path, Dir.pwd)
        relative_path = Pathname.new(absolute_file_path).relative_path_from(absolute_project_dir)
        return relative_path.to_s.gsub('/','\\')
      end

      def filter_groups(file_name)
        parts = filter_group(file_name).split('\\')
        prefixes = (0...parts.count).collect do |i|
          parts[0..i].join('\\')
        end
      end

      def filter_group(file_name)
        file_name = file_name.
          gsub('include/fonts/','').
          gsub('include/gui/','').
          gsub('include/gui_generated/','').
          gsub('images/include','images'). # images are placed differently
          gsub('include/texts/','').
          gsub('src/','')
        parts = file_name.split('/')
        parts.pop # remove file
        (parts.unshift (File.extname(file_name) == ".hpp") ? "Header Files" : "Source Files").join('\\')
      end

      def config_file
        "config/msvs/Application.props"
      end

      def config_file_contents
        File.read(config_file, :encoding=> 'utf-8' )
      end

      def config_file_xml
        @config_file_xml ||= Nokogiri::XML(config_file_contents) do |config|
          config.default_xml.noblanks
        end
      end

      def filters_file
        "#{project_file}.filters"
      end

      def project_file_contents
        File.read(project_file, :encoding=> 'utf-8' )
      end

      def filters_file_contents
        File.read(filters_file, :encoding => 'utf-8')
      end

      def project_file_xml
        @project_file_xml ||= Nokogiri::XML(project_file_contents) do |config|
          config.default_xml.noblanks
        end
      end

      def filters_file_xml
        @filters_file_xml ||= Nokogiri::XML(filters_file_contents) do |config|
          config.default_xml.noblanks
        end
      end

      def project_dir
        @project_dir ||= File.dirname(project_file)
      end

      def output_project_files
        [
          [ project_file, project_file_xml, project_file_contents ],
          [ filters_file, filters_file_xml, filters_file_contents ],
          [ config_file,  config_file_xml, config_file_contents ]
        ].each do |file, xml, contents|
          output = xml.to_xml(:indent => 2)
          if(output != contents)
            File.open(file, 'wb') { |f| f.write(output) }
          end
        end
      end

      def abort(msg)
        STDERR.puts msg; Kernel.abort
      end
    end
  end
end
