require 'touchgfx/generators/base'
require 'nokogiri'
require 'pathname'

module TouchGFX
  module Generators
    class SW4ProjectUpdater
      attr_reader :project_file
      attr_reader :touchgfx_path
      attr_reader :bpp

      # https://help.eclipse.org/luna/topic/org.eclipse.platform.doc.isv/reference/misc/project_description_file.html
      def initialize(project_file, touchgfx_path, touchgfx_env, config, gui_group_name = 'gui', generated_group_name = 'generated')
        @project_file = project_file
        @touchgfx_path = touchgfx_path
        @touchgfx_env = touchgfx_env
        @bpp = config.bpp
        @gui_group_name = gui_group_name
        @generated_group_name = generated_group_name
      end

      def run
        check_project_node
        ensure_source_nodes
        update_includes
        update_defines
        update_library
        output_project_files
      end

      private

      def check_project_node
        project_file_xml.at_xpath('./projectDescription') ||
          abort("no <projectDescription> node in #{project_file}")
        cproject_file_xml.at_xpath('./cproject') ||
          abort("no <cproject> node in #{cproject_file}")
      end

      def ensure_source_nodes
        ui_files = Dir.glob('{generated,gui}/**/*.{cpp,c}').inject({}) do |m,f| 
          m[project_path_to(f)] = f; m 
        end

        linkedResources = project_file_xml.at_xpath('./projectDescription/linkedResources')
        links = project_file_xml.xpath('./projectDescription/linkedResources/link')
        links.each do |l|
          location = l.at_xpath('./location') || l.at_xpath('./locationURI')
          if(location)
            file_name = location.content
            if( file_name.start_with?( project_path_to('generated') ) || file_name.start_with?( project_path_to('gui') ) )
              if( ui_files.has_key?( file_name ) )
                ui_files[file_name] = 0;
              else
                l.remove()
              end
            end
          end
        end

        files_to_add = ui_files.select { |k,v| v != 0 }.collect { |k,v| v }
        files_to_add.each do |file_name|
          link = Nokogiri::XML::Node.new "link", project_file_xml
          name_node = Nokogiri::XML::Node.new 'name', project_file_xml
          type_node = Nokogiri::XML::Node.new 'type', project_file_xml
          location_node = Nokogiri::XML::Node.new 'location', project_file_xml
          name_node.content = file_name.gsub('/src/','/').split('/')[1..-1].join('/')
          name_node.content = File.join(file_name.start_with?('gui/') ? @gui_group_name : @generated_group_name, name_node.content)
          type_node.content = '1'
          location_node.content = project_path_to(file_name)
          link.add_child name_node
          link.add_child type_node
          link.add_child location_node
          linkedResources.add_child link
        end
      end

      def update_includes
        includes = cproject_file_xml.xpath('//option/listOptionValue')
        includes.each do |inc|
          inc['value'] = inc['value'].gsub( /^.*?touchgfx\/framework\/include/, cproject_path_to( File.join( touchgfx_path, 'framework', 'include' ) ) )
        end
      end

      def update_defines
        define_groups = cproject_file_xml.xpath('//option[@valueType="definedSymbols"]')
        define_groups.each do |dg|
          defines = dg.xpath('./listOptionValue')
          use_bpp = defines.find { |d| d['value'].match(/^USE_BPP=\d+$/) } || begin
            lov = Nokogiri::XML::Node.new "listOptionValue", cproject_file_xml
            lov['builtIn'] = "false"
            lov['value'] = ""
            dg.add_child lov
          end
          use_bpp['value'] = "USE_BPP=#{bpp}"
        end
      end

      def update_library
        library = cproject_file_xml.xpath('//option/listOptionValue')
        library.each do |inc|
          inc['value'] = inc['value'].gsub( /^.*?touchgfx\/lib/, cproject_path_to( File.join( touchgfx_path, 'lib' ) ) )
        end
      end

      # PARENT-2-PROJECT_LOC means ../../ from project location
      def project_path_to(file_path)
        absolute_file_path = File.absolute_path(file_path, Dir.pwd)
        project_dir = File.dirname(project_file)
        absolute_project_dir = File.absolute_path(project_dir, Dir.pwd)
        relative_path = Pathname.new(absolute_file_path).relative_path_from(Pathname.new(absolute_project_dir))
        number_of_dotdots, remaining_path = 10.downto(0).collect { |d| [d, relative_path.to_s.match(/^(\.\.\/){#{d}}(.*)/) ] }.find { |d,m| !m.nil? }
        (number_of_dotdots > 0) ? File.join("PARENT-#{number_of_dotdots}-PROJECT_LOC", remaining_path[2]) : relative_path
      end

      def cproject_path_to(file_path)
        absolute_file_path = File.absolute_path(file_path, Dir.pwd)
        project_dir = File.dirname(cproject_file)
        absolute_project_dir = File.absolute_path(project_dir, Dir.pwd)
        relative_path = Pathname.new(absolute_file_path).relative_path_from(Pathname.new(absolute_project_dir)).to_s
        File.join('..', relative_path)
      end

      def cproject_file
        project_file.sub(".project", ".cproject")
      end

      def project_file_contents
        File.read(project_file, :encoding=>'utf-8')
      end

      def cproject_file_contents
        File.read(cproject_file, :encoding=>'utf-8')
      end

      def project_file_xml
        @project_file_xml ||= Nokogiri::XML(project_file_contents) do |config|
          config.default_xml.noblanks
        end
      end

      def cproject_file_xml
        @cproject_file_xml ||= Nokogiri::XML(cproject_file_contents) do |config|
          config.default_xml.noblanks
        end
      end

      def output_project_files
        [
          [ project_file, project_file_xml, project_file_contents ],
          [ cproject_file, cproject_file_xml, cproject_file_contents ]
        ].each do |file, xml, contents|
          output = xml.to_xml(:indent => 2)
          if(output != contents)
            File.open(file, 'wb') { |f| f.write(output) }
          end
        end
      end

      def file_order
        proc { |word| word.downcase }
      end

      def abort(msg)
        STDERR.puts msg; Kernel.abort
      end
    end
  end
end
